unit XCanvasObjects;

interface

uses Windows, XClasses, XCanvases, XGraphics;

type
   AObjectData = ^XObjectData;
   XObjectData = record
      Handle : Integer;
      Hash : Integer;
      Objects : ZList;
   end;

   XCanvasObjectsManager = class
   private
      fPenList : ZList; // of XObjectData
      fBrushList : ZList; // of XObjectData
      fFontList : ZList; // of XObjectData
      procedure ObjectCreating( Sender : TObject );
      procedure ObjectChanged( Sender : TObject );
      //procedure ObjectCreated( Sender : TObject );
      function ListOfObjects( Sender : TObject ) : ZList;
      function CalculateHashCode( Sender : TObject ) : Integer;
      function ObjectsEqual( A, B : TObject ) : Boolean;
   protected
   public
      constructor Create;
      destructor Destroy; override;
   end;

var CanvasObjectsManager : XCanvasObjectsManager;

procedure UseCanvasObjectsManager;

implementation

procedure UseCanvasObjectsManager;
begin
   if CanvasObjectsManager <> nil then Exit;
   CanvasObjectsManager := XCanvasObjectsManager.Create;
end;

constructor XCanvasObjectsManager.Create;
begin
   inherited;
   if GlobalCanvasManager = nil then
      GlobalCanvasManager := XCanvasManager.Create;
   GlobalCanvasManager.OnObjectCreating := ObjectCreating;
   //GlobalCanvasManager.OnObjectCreated := ObjectCreated;
   GlobalCanvasManager.OnObjectChanged := ObjectChanged;
   fPenList := ZList.Create;
   fBrushList := ZList.Create;
   fFontList := ZList.Create;
end;

destructor XCanvasObjectsManager.Destroy;
   procedure FreeObjectsList( List : ZList );
   var Data : AObjectData;
       I : Integer;
   begin
      for I := 0 to List.Count - 1 do
      begin
         Data := List.Items[ I ];
         Data.Objects.Free;
         FreeMem( Data );
      end;
      List.Free;
   end;
begin
   FreeObjectsList( fPenList );
   FreeObjectsList( fBrushList );
   FreeObjectsList( fFontList );
   inherited;
end;

procedure XCanvasObjectsManager.ObjectCreating( Sender : TObject );
var List : ZList;
    Hash : Integer;
    I : Integer;
    Data : AObjectData;
begin
   List := ListOfObjects( Sender );
   Hash := CalculateHashCode( Sender );
   for I := 0 to List.Count - 1 do
   begin
      Data := List.Items[ I ];
      if Data.Hash = Hash then
      if ObjectsEqual( Data.Objects.Items[ 0 ], Sender ) then
      begin
         Data.Objects.Add( Sender );
         (Sender as XGraphicObject).AssignHandle( Data.Handle );
         Exit;
      end;
   end;
end;

procedure XCanvasObjectsManager.ObjectChanged( Sender : TObject );
var List : ZList;
    Hash : Integer;
    I, J : Integer;
    Data : AObjectData;
begin
   List := ListOfObjects( Sender );
   Hash := CalculateHashCode( Sender );
   for I := 0 to List.Count - 1 do
   begin
      Data := List.Items[ I ];
      if Data.Hash = Hash then
      if ObjectsEqual( Data.Objects.Items[ 0 ], Sender ) then
      begin
         J := Data.Objects.IndexOf( Sender );
         if J >= 0 then
         begin
            Data.Objects.Delete( J );
            (Sender as XGraphicObject).ReleaseHandle;
            if Data.Objects.Count = 0 then
            begin
               DeleteObject( Data.Handle );
               List.Delete( I );
            end;
         end;
         Exit;
      end;
   end;
end;

function XCanvasObjectsManager.ListOfObjects( Sender : TObject ) : ZList;
begin
   Result := nil;
   if Sender is XPen then
      Result := fPenList
   else
   if Sender is XBrush then
      Result := fBrushList
   else
   if Sender is XFont then
      Result := fFontList;
end;

function AddToHash( Hash : Integer; const Buffer; Count : Integer ) : Integer;
assembler;
asm
@@1:
   ROL EAX, 5
   XOR AL, [ EDX ]
   INC EDX
   LOOP @@1
end;

function XCanvasObjectsManager.CalculateHashCode( Sender : TObject ) : Integer;
begin
   Result := 0;
   if not (Sender is XGraphicObject) then Exit;
   Result := AddToHash( Result, (Sender as XGraphicObject).Color, sizeof( XColor ) );
   if Sender is XPen then
   begin
      Result := AddToHash( Result, (Sender as XPen).Width, Sizeof( Integer ) );
      Result := AddToHash( Result, (Sender as XPen).Style, Sizeof( XPenStyle ) );
      Result := AddToHash( Result, (Sender as XPen).Mode, Sizeof( XPenMode ) );
   end
      else
   if Sender is XBrush then
   begin
      Result := AddToHash( Result, (Sender as XBrush).Bitmap, Sizeof( HBitmap ) );
      Result := AddToHash( Result, (Sender as XBrush).Style, Sizeof( XBrushStyle ) );
   end
      else
   if Sender is XFont then
   begin
      Result := AddToHash( Result, (Sender as XFont).Height, Sizeof( Integer ) );
      Result := AddToHash( Result, (Sender as XFont).Pitch, Sizeof( XFontPitch ) );
      Result := AddToHash( Result, (Sender as XFont).Style, Sizeof( XFontStyle ) );
      Result := AddToHash( Result, (Sender as XFont).Charset, Sizeof( XFontCharset ) );
      Result := AddToHash( Result, (Sender as XFont).Orientation, Sizeof( Integer ) );
      Result := AddToHash( Result, PChar( (Sender as XFont).Name )^, Length( (Sender as XFont).Name ) );
   end;
end;

function XCanvasObjectsManager.ObjectsEqual( A, B : TObject ) : Boolean;
var APen, BPen : XPen;
    ABrush, BBrush : XBrush;
    AFont, BFont : XFont;
begin
   Result := True;
   if A = B then Exit;
   Result := False;
   if (A is XPen) and (B is XPen) then
   begin
      APen := A as XPen; BPen := B as XPen;
      if APen.Color <> BPen.Color then Exit;
      if APen.Width <> BPen.Width then Exit;
      if APen.Style <> BPen.Style then Exit;
      if APen.Mode <> BPen.Mode then Exit;
      Result := True;
   end
      else
   if (A is XBrush) and (B is XBrush) then
   begin
      ABrush := A as XBrush; BBrush := B as XBrush;
      if ABrush.Color <> BBrush.Color then Exit;
      if ABrush.Bitmap <> BBrush.Bitmap then Exit;
      if ABrush.Style <> BBrush.Style then Exit;
      Result := True;
   end
      else
   if (A is XFont) and (B is XFont) then
   begin
      AFont := A as XFont; BFont := B as XFont;
      if AFont.Color <> BFont.Color then Exit;
      if AFont.Height <> BFont.Height then Exit;
      if AFont.Pitch <> BFont.Pitch then Exit;
      if AFont.Style <> BFont.Style then Exit;
      if AFont.Charset <> BFont.Charset then Exit;
      if AFont.Orientation <> BFont.Orientation then Exit;
      if AFont.Name <> BFont.Name then Exit;
      Result := True;
   end;
end;

end.
