unit XCanvases;

interface

uses Windows, XClasses, {XStrUtils,} XWindows, XGraphics;

type
   XCanvasStates = ( csHandleValid, csPenValid, csBrushValid, csFontValid );
   XCanvasState = Set of XCanvasStates;

   XFillStyle = (fsSurface, fsBorder);
   XFillMode = (fmAlternate, fmWinding);
   XCopyMode = Integer;

const
  cmBlackness = BLACKNESS;
  cmDstInvert = DSTINVERT;
  cmMergeCopy = MERGECOPY;
  cmMergePaint = MERGEPAINT;
  cmNotSrcCopy = NOTSRCCOPY;
  cmNotSrcErase = NOTSRCERASE;
  cmPatCopy = PATCOPY;
  cmPatInvert = PATINVERT;
  cmPatPaint = PATPAINT;
  cmSrcAnd = SRCAND;
  cmSrcCopy = SRCCOPY;
  cmSrcErase = SRCERASE;
  cmSrcInvert = SRCINVERT;
  cmSrcPaint = SRCPAINT;
  cmWhiteness = WHITENESS;

type
   XOnTextArea = procedure( Sender : TObject; var Size : TSize; var P0 : TPoint )
                            of object;

   XCanvas = class
   private
      fHandle : HDC;
      fPenPos : TPoint;
      fPen : XPen;
      fBrush : XBrush;
      fFont : XFont;
      fState : XCanvasState;
      fCopyMode : XCopyMode;
      fOnChanging : XOnEvent;
      fOnChanged : XOnEvent;
      procedure SetHandle( Value : HDC );
      procedure DeselectHandles;
      procedure RequiredState( ReqState : XCanvasState );
      function GetPenPos : TPoint;
      procedure SetPenPos( Value : TPoint );
      procedure CreatePen;
      procedure CreateBrush;
      procedure CreateFont;
      procedure ObjectChanged( Sender : TObject );
      procedure Changing;
      procedure Changed;
   protected
   public
      constructor Create;
      destructor Destroy; override;
      property Handle : HDC read fHandle write SetHandle;
      property PenPos : TPoint read GetPenPos write SetPenPos;
      property Pen : XPen read fPen;
      property Brush : XBrush read fBrush;
      property Font : XFont read fFont;
      procedure Arc(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Integer);
      procedure Chord(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Integer);
      procedure DrawFocusRect(const Rect: TRect);
      procedure Ellipse(X1, Y1, X2, Y2: Integer);
      procedure FillRect(const Rect: TRect);
      procedure FloodFill(X, Y: Integer; Color: XColor; FillStyle: XFillStyle);
      procedure FrameRect(const Rect: TRect);
      procedure MoveTo( X, Y : Integer );
      procedure LineTo( X, Y : Integer );
      procedure Pie(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Integer);
      procedure Polygon(const Points: array of TPoint);
      procedure Polyline(const Points: array of TPoint);
      procedure Rectangle(X1, Y1, X2, Y2: Integer);
      procedure RoundRect(X1, Y1, X2, Y2, X3, Y3: Integer);
      procedure TextOut(X, Y: Integer; const Text: String);
      procedure TextRect(Rect: TRect; X, Y: Integer; const Text: string);
      function TextExtent(const Text: string): TSize;
      procedure TextArea( const Text : String; var Sz : TSize; var P0 : TPoint );
      function TextWidth(const Text: string): Integer;
      function TextHeight(const Text: string): Integer;
      property CopyMode : XCopyMode read fCopyMode write fCopyMode;
      procedure CopyRect( DstRect : TRect; SrcCanvas : XCanvas; SrcRect : TRect );
      property OnChanging : XOnEvent read fOnChanging write fOnChanging;
      property OnChanged : XOnEvent read fOnChanged write fOnChanged;
   end;

   XCanvasManager = class
   private
      fOnCreateCanvas : XOnEvent;
      fOnDestroyCanvas : XOnEvent;
      fOnObjectChanged : XOnEvent;
      fOnObjectCreating : XOnEvent;
      fOnObjectCreated : XOnEvent;
      fOnTextArea : XOnTextArea;
      procedure GraphicObjectCreating( Sender : TObject );
      procedure GraphicObjectCreated( Sender : TObject );
      procedure WindowCreateCanvas( Sender : TObject );
      procedure WindowDestroyCanvas( Sender : TObject );
      procedure WindowBeginPaint( Sender : TObject; DC : HDC );
      procedure WindowEndPaint( Sender : TObject; DC : HDC );
   public
      constructor Create;
      property OnCreateCanvas : XOnEvent read fOnCreateCanvas write fOnCreateCanvas;
      property OnDestroyCanvas : XOnEvent read fOnDestroyCanvas write fOnDestroyCanvas;
      property OnObjectChanged : XOnEvent read fOnObjectChanged write fOnObjectChanged;
      property OnObjectCreating : XOnEvent read fOnObjectCreating write fOnObjectCreating;
      property OnObjectCreated : XOnEvent read fOnObjectCreated write fOnObjectCreated;
      property OnTextArea : XOnTextArea read fOnTextArea write fOnTextArea;
   end;

var GlobalCanvasManager : XCanvasManager;

function WindowCanvas( Window : XWindow ) : XCanvas;

implementation

var
  StockPen: HPEN;
  StockBrush: HBRUSH;
  StockFont: HFONT;

constructor XCanvasManager.Create;
begin
   inherited;
   if GlobalGraphicsManager = nil then
   begin
      GlobalGraphicsManager := XGraphicsManager.Create;
   end;
   GlobalGraphicsManager.OnObjectCreating := GraphicObjectCreating;
   GlobalGraphicsManager.OnObjectCreated := GraphicObjectCreated;
   GlobalWindowsManager.OnCreateWindow := WindowCreateCanvas;
   GlobalWindowsManager.OnDestroyWindow := WindowDestroyCanvas;
   GlobalWindowsManager.OnBeginPaint := WindowBeginPaint;
   GlobalWindowsManager.OnEndPaint := WindowEndPaint;
end;

procedure XCanvasManager.GraphicObjectCreating( Sender : TObject );
begin
   if Assigned( fOnObjectCreating ) then
      fOnObjectCreating( Sender );
end;

procedure XCanvasManager.GraphicObjectCreated( Sender : TObject );
begin
   if Assigned( fOnObjectCreated ) then
      fOnObjectCreated( Sender );
end;

procedure XCanvasManager.WindowCreateCanvas( Sender : TObject );
begin
   (Sender as XWindow).Canvas := XCanvas.Create;
end;

procedure XCanvasManager.WindowDestroyCanvas( Sender : TObject );
begin
   XCanvas( (Sender as XWindow).Canvas ).Free;
end;

procedure XCanvasManager.WindowBeginPaint( Sender : TObject; DC : HDC );
begin
   XCanvas( (Sender as XWindow).Canvas ).Handle := DC;
end;

procedure XCanvasManager.WindowEndPaint( Sender : TObject; DC : HDC );
begin
   if XCanvas( (Sender as XWindow).Canvas ).fHandle = DC then
      XCanvas( (Sender as XWindow).Canvas ).Handle := 0;
end;

function WindowCanvas( Window : XWindow ) : XCanvas;
begin
   Result := XCanvas( Window.Canvas );
end;

constructor XCanvas.Create;
begin
   {
   if GlobalCanvasManager = nil then
      GlobalCanvasManager := XCanvasManager.Create;
   }
   inherited;
   fPen := XPen.Create;
   fPen.OnChange := ObjectChanged;
   fBrush := XBrush.Create;
   fBrush.Color := clBtnFace;
   fBrush.OnChange := ObjectChanged;
   fFont := XFont.Create;
   fFont.OnChange := ObjectChanged;
   if Assigned( GlobalCanvasManager.fOnCreateCanvas ) then
      GlobalCanvasManager.OnCreateCanvas( Self );
end;

destructor XCanvas.Destroy;
begin
   Handle := 0;
   fPen.Free;
   fBrush.Free;
   fFont.Free;
   if GlobalCanvasManager <> nil then
   if Assigned( GlobalCanvasManager.fOnDestroyCanvas ) then
      GlobalCanvasManager.OnDestroyCanvas( Self );
   inherited;
end;

procedure XCanvas.SetHandle( Value : HDC );
begin
   if fHandle = Value then Exit;
   if fHandle <> 0 then
   begin
      fPenPos := GetPenPos;
      DeselectHandles;
      fHandle := 0;
      Exclude( fState, csHandleValid );
   end;
   if Value <> 0 then
   begin
      Include( fState, csHandleValid );
      fHandle := Value;
      SetPenPos( fPenPos );
   end;
end;

procedure XCanvas.DeselectHandles;
begin
   if (fHandle <> 0) and
      (fState - [ csPenValid, csBrushValid, csFontValid ] <> fState) then
   begin
      SelectObject( fHandle, StockPen );
      SelectObject( fHandle, StockBrush );
      SelectObject( fHandle, StockFont );
      fState := fState - [ csPenValid, csBrushValid, csFontValid ];
   end;
end;

procedure XCanvas.RequiredState( ReqState : XCanvasState );
var
  NeededState: XCanvasState;
begin
  NeededState := ReqState - fState;
  if NeededState <> [] then
  begin
    if csHandleValid in NeededState then
    begin
      //CreateHandle;
      if fHandle = 0 then Exit;
      //if fHandle = 0 then
      //   NeededState := NeededState - [ csHandleValid ];
    end;
    if csFontValid in NeededState then
       CreateFont;
    if csPenValid in NeededState then
    begin
      CreatePen;
      if Pen.Style in [psDash, psDot, psDashDot, psDashDotDot] then
        Include(NeededState, csBrushValid);
    end;
    if csBrushValid in NeededState then
       CreateBrush;
    fState := fState + NeededState;
  end;
end;

procedure XCanvas.ObjectChanged( Sender : TObject );
begin
   DeselectHandles;
   if Assigned( GlobalCanvasManager.fOnObjectChanged ) then
      GlobalCanvasManager.fOnObjectChanged( Sender );
end;

function XCanvas.GetPenPos : TPoint;
begin
   RequiredState( [csHandleValid] );
   Windows.GetCurrentPositionEx( fHandle, @fPenPos );
   Result := fPenPos;
end;

procedure XCanvas.SetPenPos( Value : TPoint );
begin
   MoveTo( Value.x, Value.y );
end;

procedure XCanvas.CreatePen;
const
  PenModes: array[ XPenMode ] of Word =
    (R2_BLACK, R2_WHITE, R2_NOP, R2_NOT, R2_COPYPEN, R2_NOTCOPYPEN, R2_MERGEPENNOT,
     R2_MASKPENNOT, R2_MERGENOTPEN, R2_MASKNOTPEN, R2_MERGEPEN, R2_NOTMERGEPEN,
     R2_MASKPEN, R2_NOTMASKPEN, R2_XORPEN, R2_NOTXORPEN);
begin
   SelectObject( fHandle, Pen.Handle );  //4//
   SetROP2( fHandle, PenModes[ Pen.Mode ] );
end;

procedure XCanvas.CreateBrush;
begin
  UnrealizeObject( Brush.Handle );
  SelectObject( fHandle, Brush.Handle );
  if Brush.Style = bsSolid then
  begin
     SetBkColor( fHandle, Color2RGB( Brush.Color ) );
     SetBkMode( fHandle, OPAQUE );
  end
  else
  begin
     { Win95 doesn't draw brush hatches if bkcolor = brush color }
     { Since bkmode is transparent, nothing should use bkcolor anyway }
     SetBkColor( fHandle, not Color2RGB( Brush.Color ) );
     SetBkMode( fHandle, TRANSPARENT );
  end;
end;

procedure XCanvas.CreateFont;
begin
   SelectObject( fHandle, Font.Handle ); //4//
   SetTextColor( fHandle, Color2RGB( Font.Color ) );
end;

procedure XCanvas.MoveTo( X, Y : Integer );
begin
   RequiredState( [csHandleValid] );
   Windows.MoveToEx( fHandle, X, Y, nil );
end;

procedure XCanvas.Arc(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Integer);
begin
  Changing;
  RequiredState([csHandleValid, csPenValid]);
  Windows.Arc(FHandle, X1, Y1, X2, Y2, X3, Y3, X4, Y4);
  Changed;
end;

procedure XCanvas.Chord(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Integer);
begin
  Changing;
  RequiredState([csHandleValid, csPenValid, csBrushValid]);
  Windows.Chord(FHandle, X1, Y1, X2, Y2, X3, Y3, X4, Y4);
  Changed;
end;

procedure XCanvas.DrawFocusRect(const Rect: TRect);
begin
  Changing;
  RequiredState([csHandleValid, csBrushValid]);
  Windows.DrawFocusRect(FHandle, Rect);
  Changed;
end;

procedure XCanvas.Ellipse(X1, Y1, X2, Y2: Integer);
begin
  Changing;
  RequiredState([csHandleValid, csPenValid, csBrushValid]);
  Windows.Ellipse(FHandle, X1, Y1, X2, Y2);
  Changed;
end;

procedure XCanvas.FillRect(const Rect: TRect);
begin
  Changing;
  RequiredState([csHandleValid, csBrushValid]);
  Windows.FillRect(FHandle, Rect, Brush.Handle); //4//
  Changed;
end;

procedure XCanvas.FloodFill(X, Y: Integer; Color: XColor;
  FillStyle: XFillStyle);
const
  FillStyles: array[XFillStyle] of Word =
    (FLOODFILLSURFACE, FLOODFILLBORDER);
begin
  Changing;
  RequiredState([csHandleValid, csBrushValid]);
  Windows.ExtFloodFill(FHandle, X, Y, Color, FillStyles[FillStyle]);
  Changed;
end;

procedure XCanvas.FrameRect(const Rect: TRect);
begin
  Changing;
  RequiredState([csHandleValid, csBrushValid]);
  Windows.FrameRect(FHandle, Rect, Brush.Handle); //4//
  Changed;
end;

procedure XCanvas.LineTo( X, Y : Integer );
begin
   Changing;
   RequiredState( [csHandleValid, csPenValid]);
   Windows.LineTo( fHandle, X, Y );
   Changed;
end;

procedure XCanvas.Pie(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Integer);
begin
  Changing;
  RequiredState([csHandleValid, csPenValid, csBrushValid]);
  Windows.Pie( fHandle, X1, Y1, X2, Y2, X3, Y3, X4, Y4);
  Changed;
end;

type
  PPoints = ^TPoints;
  TPoints = array[0..0] of TPoint;

procedure XCanvas.Polygon(const Points: array of TPoint);
begin
  Changing;
  RequiredState([csHandleValid, csPenValid, csBrushValid]);
  Windows.Polygon( fHandle, PPoints(@Points)^, High(Points) + 1);
  Changed;
end;

procedure XCanvas.Polyline(const Points: array of TPoint);
begin
  Changing;
  RequiredState([csHandleValid, csPenValid, csBrushValid]);
  Windows.Polyline( fHandle, PPoints(@Points)^, High(Points) + 1);
  Changed;
end;

procedure XCanvas.Rectangle(X1, Y1, X2, Y2: Integer);
begin
  Changing;
  RequiredState([csHandleValid, csBrushValid, csPenValid]);
  Windows.Rectangle( fHandle, X1, Y1, X2, Y2);
  Changed;
end;

procedure XCanvas.RoundRect(X1, Y1, X2, Y2, X3, Y3: Integer);
begin
  Changing;
  RequiredState([csHandleValid, csBrushValid, csPenValid]);
  Windows.RoundRect( fHandle, X1, Y1, X2, Y2, X3, Y3);
  Changed;
end;

procedure XCanvas.TextOut(X, Y: Integer; const Text: String);
begin
  Changing;
  RequiredState([csHandleValid, csFontValid, csBrushValid]);
  Windows.TextOut(FHandle, X, Y, PChar(Text), Length(Text));
  MoveTo(X + TextWidth(Text), Y);
  Changed;
end;

procedure XCanvas.TextRect(Rect: TRect; X, Y: Integer; const Text: string);
var
  Options: Integer;
begin
  Changing;
  RequiredState([csHandleValid, csFontValid, csBrushValid]);
  Options := ETO_CLIPPED;
  if Brush.Style <> bsClear then Inc(Options, ETO_OPAQUE);
  Windows.ExtTextOut( fHandle, X, Y, Options, @Rect, PChar(Text),
    Length(Text), nil);
  Changed;
end;

function XCanvas.TextExtent(const Text: string): TSize;
var DC : HDC;
    ClearDC : Boolean;
begin
  RequiredState([csHandleValid, csFontValid]);
  Result.cX := 0;
  Result.cY := 0;
  DC := fHandle;
  ClearDC := False;
  if DC = 0 then
  begin
     ClearDC := True;
     DC := CreateCompatibleDC( 0 );
     DeselectHandles;
     SetHandle( DC );
     RequiredState( [ csHandleValid, csFontValid ] );
  end;
  Windows.GetTextExtentPoint( DC, PChar(Text), Length(Text), Result);
  if ClearDC then
  begin
     SetHandle( 0 );
     DeleteDC( DC );
  end;
end;

function XCanvas.TextWidth(const Text: string): Integer;
begin
  Result := TextExtent(Text).cX;
end;

function XCanvas.TextHeight(const Text: string): Integer;
begin
  Result := TextExtent(Text).cY;
end;

procedure XCanvas.CopyRect( DstRect : TRect; SrcCanvas : XCanvas; SrcRect : TRect );
begin
  Changing;
  RequiredState([csHandleValid, csFontValid, csBrushValid]);
  SrcCanvas.RequiredState([csHandleValid, csBrushValid]);
  StretchBlt( fHandle, DstRect.Left, DstRect.Top, DstRect.Right - DstRect.Left,
    DstRect.Bottom - DstRect.Top, SrcCanvas.Handle, SrcRect.Left, SrcRect.Top,
    SrcRect.Right - SrcRect.Left, SrcRect.Bottom - SrcRect.Top, CopyMode);
  Changed;
end;

procedure XCanvas.Changing;
begin
   if Assigned( fOnChanging ) then
      fOnChanging( Self );
end;

procedure XCanvas.Changed;
begin
   if Assigned( fOnChanged ) then
      fOnChanged( Self );
end;

procedure XCanvas.TextArea( const Text : String; var Sz : TSize; var P0 : TPoint );
begin
   Sz := TextExtent( Text );
   P0.x := 0; P0.y := 0;
   if Assigned( GlobalCanvasManager.fOnTextArea ) then
      GlobalCanvasManager.fOnTextArea( Self, Sz, P0 );
end;

initialization
   if GlobalCanvasManager = nil then
      GlobalCanvasManager := XCanvasManager.Create;
   StockPen := GetStockObject(BLACK_PEN);
   StockBrush := GetStockObject(HOLLOW_BRUSH);
   StockFont := GetStockObject(SYSTEM_FONT);

finalization
   GlobalCanvasManager.Free;
   GlobalCanvasManager := nil;

end.
