unit XVisuals;

interface

uses Windows, XClasses, XGraphics;

type
   XVisual = class( XClass )
   private
      fEnabled : Boolean;
      fVisible : Boolean;
      fBoundsRect : TRect;
      fColor : XColor;
      function GetLeft : Integer;
      procedure SetLeft( Value : Integer );
      function GetTop : Integer;
      procedure SetTop( Value : Integer );
      function GetWidth : Integer;
      procedure SetWidth( Value : Integer );
      function GetHeight : Integer;
      procedure SetHeight( Value : Integer );
      function GetPosition : TPoint;
      procedure SetPosition( Value : TPoint );
   protected
      function GetEnabled : Boolean; virtual;
      procedure SetEnabled( Value : Boolean ); virtual;
      function GetVisible : Boolean; virtual;
      procedure SetVisible( Value : Boolean ); virtual;
      function GetBoundsRect : TRect; virtual;
      procedure SetBoundsRect( Value : TRect ); virtual;
      procedure DoPaint( DC : HDC; Rect : TRect ); dynamic;
      procedure PaintChilds( DC : HDC; Rect : TRect ); dynamic;
      procedure PaintErase( DC : HDC; Rect : TRect );
      function GetWindowHandle : HWnd; dynamic;
      procedure SetColor( Value : XColor ); virtual;
      property Color : XColor read fColor write SetColor;
      function GetWndUpdRgn : HRgn; dynamic;
      procedure ExcludeUpdRect( DC : HDC; Rect : TRect ); dynamic;
      procedure ExcludeUpdRgn( DC : HDC; Rgn : HRgn); dynamic;
      function CreateWindow : Boolean; dynamic;
    function GetParentWnd( NeedHandle : Boolean ): HWnd; dynamic;
      procedure CreateChildWindows;
   public
      constructor Create( AParent : XClass ); override;
      property Enabled : Boolean read GetEnabled write SetEnabled;
      property Visible : Boolean read GetVisible write SetVisible;
      property BoundsRect : TRect read GetBoundsRect write SetBoundsRect;
      property Left : Integer read GetLeft write SetLeft;
      property Top : Integer read GetTop write SetTop;
      property Width : Integer read GetWidth write SetWidth;
      property Height : Integer read GetHeight write SetHeight;
      property Position : TPoint read GetPosition write SetPosition;
      function ClientRect : TRect; dynamic; abstract;
      function ControlRect : TRect; dynamic;
      function ElementAtPos( X, Y : Integer; IgnoreDisabled : Boolean ) : XVisual;
      procedure Invalidate; dynamic; abstract;
      function Windowed : Boolean; dynamic; abstract;
      function WndProc( var Msg : TMsg ) : Integer; virtual;
   end;

function GetWindowHandle( V : XVisual ) : HWnd;
function CreateWindow( V : XVisual ) : Boolean;

implementation

constructor XVisual.Create( AParent : XClass );
begin
   inherited;
   fVisible := True;
   fEnabled := True;
   if (AParent <> nil) and (AParent is XVisual) then
      fColor := (AParent as XVisual).fColor
   else
      fColor := clBtnFace;
end;

function XVisual.GetEnabled : Boolean;
begin
   Result := fEnabled;
end;

procedure XVisual.SetEnabled( Value : Boolean );
begin
   fEnabled := Value;
end;

function XVisual.GetVisible : Boolean;
begin
   Result := fVisible;
end;

procedure XVisual.SetVisible( Value : Boolean );
begin
   fVisible := Value;
end;

function XVisual.GetBoundsRect : TRect;
begin
   Result := fBoundsRect;
end;

procedure XVisual.SetBoundsRect( Value : TRect );
begin
   fBoundsRect := Value;
end;

function XVisual.GetLeft : Integer;
begin
   Result := BoundsRect.Left;
end;

procedure XVisual.SetLeft( Value : Integer );
var R : TRect;
begin
   R := fBoundsRect;
   R.Left := Value;
   R.Right := Value + Width;
   SetBoundsRect( R );
end;

function XVisual.GetTop : Integer;
begin
   Result := BoundsRect.Top;
end;

procedure XVisual.SetTop( Value : Integer );
var R : TRect;
begin
   R := fBoundsRect;
   R.Top := Value;
   R.Bottom := Value + Height;
   SetBoundsRect( R );
end;

function XVisual.GetWidth : Integer;
begin
   Result := BoundsRect.Right - BoundsRect.Left;
end;

procedure XVisual.SetWidth( Value : Integer );
var R : TRect;
begin
   R := fBoundsRect;
   R.Right := R.Left + Value;
   SetBoundsRect( R );
end;

function XVisual.GetHeight : Integer;
begin
   Result := BoundsRect.Bottom - BoundsRect.Top;
end;

procedure XVisual.SetHeight( Value : Integer );
var R : TRect;
begin
   R := fBoundsRect;
   R.Bottom := R.Top + Value;
   SetBoundsRect( R );
end;

function XVisual.GetPosition : TPoint;
begin
   Result := BoundsRect.TopLeft;
end;

procedure XVisual.SetPosition( Value : TPoint );
var R : TRect;
begin
   R.TopLeft := Value;
   R.Right := R.Left + Width;
   R.Bottom := R.Top + Height;
   BoundsRect := R;
end;

function XVisual.ControlRect : TRect;
var C : XClass;
    V : XVisual;
    R : TRect;
begin
   Result := BoundsRect;
   C := Parent;
   if C <> nil then
   begin
      if not( C is XVisual ) then Exit;
      V := C as XVisual;
      if V.Windowed then Exit;
      R := V.ControlRect;
      OffsetRect( Result, R.Left, R.Top );
      R := V.ClientRect;
      OffsetRect( Result, R.Left, R.Top );
   end;
end;

function XVisual.ElementAtPos( X, Y : Integer;
                                   IgnoreDisabled : Boolean ) : XVisual;
var I : Integer;
    C : XClass;
    V : XVisual;
    R : TRect;
begin
   Result := nil;
   R := ClientRect;
   X := X - R.Left;
   Y := Y - R.Top;
   for I := ChildsCount - 1 downto 0 do
   begin
      C := Childs[ I ];
      if C is XVisual then
      begin
         V := C as XVisual;
         if V.Visible then
         if (not IgnoreDisabled) or IgnoreDisabled and V.Enabled then
         if (V.Left <= X) and (V.Left + V.Width > X) then
         if (V.Top <= Y) and (V.Top + V.Height > Y) then
         begin
            Result := V;
            Exit;
         end;
      end;
   end;
end;

procedure XVisual.DoPaint( DC : HDC; Rect : TRect );
begin
   PaintChilds( DC, Rect );
   PaintErase( DC, Rect );
end;

procedure XVisual.PaintChilds( DC : HDC; Rect : TRect );
var I : Integer;
    C : XClass;
    V : XVisual;
    R : TRect;
    Rgn : HRgn;
    Intersect : Integer;
    UpdRgn : HRgn;
begin
   for I := ChildsCount - 1 downto 0 do
   begin
      C := Childs[ I ];
      if C is XVisual then
      begin
         V := C as XVisual;
         if not V.Visible then continue;
         R := V.ControlRect;
         UpdRgn := GetWndUpdRgn;
         if UpdRgn <> 0 then
         begin
            Rgn := CreateRectRgn( R.Left, R.Top, R.Right, R.Bottom );
            Intersect := CombineRgn( Rgn, Rgn, UpdRgn, RGN_AND );
            DeleteObject( Rgn );
         end
            else
            Intersect := SIMPLEREGION;
         if (Intersect = SIMPLEREGION) or (Intersect = COMPLEXREGION) then
            V.DoPaint( DC, R );
      end;
   end;
end;

procedure XVisual.PaintErase( DC : HDC; Rect : TRect );
var B : HBrush;
    R : TRect;
begin
   R := Rect;
   B := CreateSolidBrush( Color2Rgb( Color ) );
   Windows.FillRect( DC, R, B );
   DeleteObject( B );
   ExcludeUpdRect( DC, R );
end;

procedure XVisual.SetColor( Value : XColor );
begin
   if fColor = Value then Exit;
   fColor := Value;
   Invalidate;
end;

function XVisual.GetWndUpdRgn: HRgn;
begin
   Result := 0;
end;

procedure XVisual.ExcludeUpdRect(DC: HDC; Rect: TRect);
begin
   ExcludeClipRect( DC, Rect.Left, Rect.Top, Rect.Right, Rect.Bottom );
end;

procedure XVisual.ExcludeUpdRgn(DC: HDC; Rgn: HRgn);
begin
   ExtSelectClipRgn( DC, Rgn, RGN_DIFF );
end;

function XVisual.GetWindowHandle: HWnd;
begin
   Result := 0;
end;

function XVisual.WndProc(var Msg: TMsg): Integer;
begin
   Result := DefWindowProc( Msg.hwnd, Msg.message, Msg.WParam, Msg.LParam );
end;

function XVisual.CreateWindow : Boolean;
begin
  Result := False;
  CreateChildWindows;
end;

function GetWindowHandle( V : XVisual ) : HWnd;
begin
   Result := V.GetWindowHandle;
end;

function CreateWindow( V : XVisual ) : Boolean;
begin
   Result := V.CreateWindow;
end;

function XVisual.GetParentWnd( NeedHandle : Boolean ): HWnd;
var C : XClass;
    V : XVisual;
begin
   Result := 0;
   C := Parent;
   while C <> nil do
   begin
      if C is XVisual then
      begin
         V := C as XVisual;
         if V.Windowed then
         begin
            Result := XVisuals.GetWindowHandle( V );
            if Result = 0 then
            begin
               if not NeedHandle then Exit;
               XVisuals.CreateWindow( V );
               Result := XVisuals.GetWindowHandle( V );
            end;
            if Result = 0 then Exit;
            break;
         end;
      end;
      C := C.Parent;
   end;
end;

procedure XVisual.CreateChildWindows;
var I : Integer;
    C : XClass;
    V : XVisual;
begin
  for I := ChildsCount - 1 downto 0 do
  begin
     C := Childs[ I ];
     if C is XVisual then
     begin
        V := C as XVisual;
        V.CreateWindow;
     end;
  end;
end;

end.
